from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from functools import partial
from typing import Any, Literal, override

from greyhorse.maybe import Just, Maybe, Nothing
from greyhorse.utils.types import TypeWrapper


class Operator[T](TypeWrapper[T], ABC):
    @abstractmethod
    def accept(self, instance: T) -> bool | Awaitable[bool]: ...

    @abstractmethod
    def revoke(self) -> Maybe[T] | Awaitable[Maybe[T]]: ...


class AssignOperator[T](Operator[T]):
    def __init__(
        self, getter: Callable[[], Maybe[T]], setter: Callable[[Maybe[T]], Any]
    ) -> None:
        super().__init__()
        self._getter = getter
        self._setter = setter

    def _set(self, value: T) -> Literal[True]:
        self._setter(Just(value))
        return True

    def _clear(self, value: T) -> T:
        self._setter(Nothing)
        return value

    @override
    def accept(self, instance: T) -> bool:
        return self._getter().map_or_else(partial(self._set, instance), lambda _: False)

    @override
    def revoke(self) -> Maybe[T]:
        return self._getter().map(self._clear)


class AttrOperator[T](AssignOperator[T]):
    def __init__(self, instance: object, attr: str) -> None:
        if hasattr(instance, attr):
            super().__init__(partial(getattr, instance, attr), partial(setattr, instance, attr))
        else:
            raise ValueError(f'Instance {instance!r} has no attribute "{attr}"')



from dataclasses import field as dataclass_field
from dataclasses import fields as dataclass_fields
from dataclasses import make_dataclass
from typing import Any, ClassVar, Generic, NoReturn, TypeVar

from greyhorse.utils.invoke import caller_path


class Unit:
    __slots__ = ('_name', '_base', '_factory')

    def _bind(self, name: str, base: type):
        self._name = name
        self._base = base

        bases = [base]

        fields = [
            ('__orig_class__', Any, dataclass_field(default=self._base, init=True, repr=False))
        ]

        dc = make_dataclass(
            name,
            fields,
            bases=tuple(bases),
            module=f'{'.'.join(caller_path(3))}.{self._base.__name__}',
            slots=True,
            frozen=True,
            repr=False,
            match_args=False,
        )

        dc.__orig_class__ = self.__class__  # type: ignore
        dc.__match_args__ = ()  # type: ignore
        dc.__repr__ = lambda _: f'{self._base.__name__}:{self._name}'  # type: ignore
        self._factory = dc()  # type: ignore
        return self._factory

    def __repr__(self) -> str:
        return f'{self._base.__name__}:{self._name}'

    def __call__(self):
        return self._factory


class Tuple[*Ts]:
    __slots__ = ('_name', '_base', '_factory', '_fields', '_types')

    def __init__(self, *types: *Ts) -> None:
        self._types: list[TypeVar] = []
        self._fields: list[type] = []

        for t in types:
            if isinstance(t, type):
                self._fields.append(t)
            elif isinstance(t, TypeVar):
                self._types.append(t)

    def _bind(self, name: str, base: type):
        self._name = name
        self._base = base

        bases = [base]

        if self._types:
            bases += [Generic[*self._types]]  # type: ignore

        fields = list(
            {f'_{i}': arg for i, arg in enumerate(self._types + self._fields)}.items()
        )  # type: ignore
        fields.append((
            '__orig_class__',
            Any,
            dataclass_field(default=self._base, init=True, repr=False),
        ))  # type: ignore

        dc = make_dataclass(
            name,
            fields,
            bases=tuple(bases),
            module=f'{'.'.join(caller_path(3))}.{self._base.__name__}',
            slots=True,
            frozen=True,
            repr=False,
            match_args=False,
        )

        dc.__orig_class__ = self.__class__  # type: ignore
        dc.__match_args__ = tuple([f'_{i}' for i, _ in enumerate(self._types + self._fields)])  # type: ignore
        dc.__repr__ = lambda self0: self._repr_fn(self0)  # type: ignore
        self._factory = dc  # type: ignore

        def __init__(self0, *args):
            if not args:
                args = (None,)
            types = tuple(type(arg) for arg in args)
            orig_class = self._factory[*types] if self._types else self.__class__  # type: ignore
            return old_init(self0, *args, __orig_class__=orig_class)

        old_init, dc.__init__ = dc.__init__, __init__  # type: ignore
        return self._factory

    def __repr__(self) -> str:
        return (
            f'{self._base.__name__}:{self._name}('
            f'{', '.join([t.__name__ for t in self._types])})'
        )

    def _repr_fn(self, instance) -> str:
        res = []
        for field, type_ in zip(
            dataclass_fields(instance), instance.__orig_class__.__args__, strict=False
        ):
            if not field.repr:
                continue

            if v := getattr(instance, field.name, ''):
                res.append(f'{type_.__name__}: {v!r}')
            else:
                res.append(f'{type_.__name__}')

        return f'{self._base.__name__}:{self._name}({', '.join(res)})'

    def __call__(self, *args):
        return self._factory(*args)


class Struct:
    __slots__ = ('_name', '_base', '_factory', '_fields', '_values')

    def __init__(self, **kwargs) -> None:
        self._values: dict[str, Any] = {}
        self._fields: dict[str, type] = {}

        for k, v in kwargs.items():
            if isinstance(v, type | TypeVar):
                self._fields[k] = v  # type: ignore
            else:
                self._values[k] = v

    def _bind(self, name: str, base: type):
        self._name = name
        self._base = base

        bases = [base]

        fields = list(self._fields.items())
        fields.append((
            '__orig_class__',
            Any,
            dataclass_field(default=self._base, init=True, repr=False),
        ))  # type: ignore

        for k, v in self._values.items():
            fields.append((k, type(v), dataclass_field(default=v, init=False, repr=True)))  # type: ignore

        dc = make_dataclass(
            name,
            fields,
            bases=tuple(bases),
            module=f'{'.'.join(caller_path(3))}.{self._base.__name__}',
            slots=True,
            frozen=True,
            repr=False,
            match_args=False,
        )

        dc.__orig_class__ = self.__class__  # type: ignore
        dc.__match_args__ = tuple([f'{k}' for k in self._fields])  # type: ignore
        dc.__repr__ = lambda self0: self._repr_fn(self0)  # type: ignore
        self._factory = dc  # type: ignore
        return self._factory

    def __repr__(self) -> str:
        return (
            f'{self._base.__name__}:{self._name}('
            f'{', '.join([f'{n}: {t.__name__}' for n, t in self._fields.items()])})'
        )

    def _repr_fn(self, instance) -> str:
        res = []
        for field in dataclass_fields(instance):
            if not field.repr:
                continue

            if v := getattr(instance, field.name, ''):
                res.append(f'{field.name}: {type(v).__name__} = {v!r}')
            else:
                res.append(f'{field.name}: {type(v).__name__}')

        return f'{self._base.__name__}:{self._name}({', '.join(res)})'

    def __call__(self, **kwargs):
        return self._factory(**kwargs)


class Enum:
    """
    Implements algebraic enumeration class.
    """

    def __init_subclass__(cls, allow_init: bool = False, **kwargs):
        fields: dict[str, ClassVar[Unit | Tuple | Struct]] = {}  # type: ignore

        for field_name in dir(cls):
            instance = getattr(cls, field_name)

            if isinstance(instance, Unit | Tuple | Struct):
                fields[field_name] = instance
                delattr(cls, field_name)
            else:
                continue

        for field_name, instance in fields.items():
            # noinspection PyProtectedMember
            factory = instance._bind(field_name, cls)
            setattr(cls, field_name, factory)

        if fields and not allow_init:

            def __init__(*_, **__) -> NoReturn:
                raise NotImplementedError()

            cls.__init__ = __init__  # type: ignore

        return super().__init_subclass__(**kwargs)


def enum(cls: type | None = None, allow_init: bool = False):
    """
    Create algebraic enumeration from class.
    """

    def decorator(cls):
        for field_name in dir(cls):
            instance = getattr(cls, field_name)

            if isinstance(instance, Unit | Tuple | Struct):
                # noinspection PyProtectedMember
                factory = instance._bind(field_name, cls)
                setattr(cls, field_name, factory)
            else:
                continue

        if not allow_init:

            def __init__(*_, **__) -> NoReturn:
                raise NotImplementedError()

            cls.__init__ = __init__

        return cls

    if cls is None:
        return decorator

    return decorator(cls)



from __future__ import annotations

from greyhorse.app.abc.operators import AssignOperator
from greyhorse.app.abc.providers import SharedProvider
from greyhorse.app.builders.module import ModuleBuilder
from greyhorse.app.schemas.components import ModuleComponentConf, ModuleConf, ProvidersConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf
from greyhorse.maybe import Maybe, Nothing

from .common.functional import FunctionalOperator
from .root import FunctionalOperatorCtrl, FunctionalOperatorService


def __init__():
    return ModuleConf(
        enabled=True,
        can_provide=[FunctionalOperator],
        components={
            'dict': ModuleComponentConf(
                enabled=True,
                path='..module.main',
                provider_imports=[
                    ProvidersConf(
                        resource=FunctionalOperator,
                        providers=[SharedProvider[FunctionalOperator]],
                    )
                ],
                services=[SvcConf(type=FunctionalOperatorService)],
                controllers=[CtrlConf(type=FunctionalOperatorCtrl)],
            )
        },
    )


def test_module() -> None:
    module_conf = __init__()

    module_builder = ModuleBuilder(module_conf, 'tests')
    res = module_builder.create_pass()
    assert res.is_ok()
    module = res.unwrap()

    op_maybe: Maybe[FunctionalOperator] = Nothing

    def assign(value) -> None:
        nonlocal op_maybe
        op_maybe = value

    sub_operator = AssignOperator[FunctionalOperator](lambda: op_maybe, assign)

    assert module.add_operator(sub_operator)

    res = module.setup()
    assert res.is_ok()

    print('OK')

    op = op_maybe.unwrap()

    res = op.add_number(123)
    assert res.is_ok()

    res = op.get_number()
    assert res.is_ok()
    assert res.unwrap() == 123

    res = op.remove_number()
    assert res.is_ok()
    assert res.unwrap() is True

    res = op.get_number()
    assert res.is_err()
    assert res.unwrap_err() == 'Number is not initialized'

    res = module.teardown()
    assert res.is_ok()



# mypy: warn_no_return=false,disable_error_code="arg-type,has-type,misc,return-value"
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeGuard, TypeVar

from .enum import Enum, Tuple, Unit

if TYPE_CHECKING:
    from .result import Result


ExcType = TypeVar('ExcType', bound=BaseException)


class Maybe[T](Enum):
    Just = Tuple(T)
    Nothing = Unit()

    def __new__(cls, value: T | None = None):
        match cls.__name__:
            case 'Maybe':
                if value is None:
                    return cls.Nothing
                return cls.__new_just__(value)
            case 'Just':
                return cls.__new_just__(value)
            case 'Nothing':
                return super().__new__(cls)

        raise AssertionError()

    @classmethod
    def __new_just__(cls, value):
        return super().__new__(Maybe[type(value)].Just)

    def __bool__(self) -> bool:
        return self.is_just()

    def __hash__(self) -> int:
        match self:
            case Maybe.Just(v):
                return hash((True, v))
            case Maybe.Nothing:
                # A large random number is used here to avoid a hash collision with
                # something else since there is no real inner value for us to hash.
                return hash((False, 969048543980197075614658413455))

    def is_just(self) -> bool:
        """Returns true if the `Maybe` is a `Just` value."""
        return isinstance(self, Maybe[T].Just)

    is_some = is_just

    def is_just_and(self, f: Callable[[T], bool]) -> bool:
        """
        Returns true if the `Maybe` is a `Just` and the value inside of it matches a predicate.
        """
        match self:
            case Maybe.Just(v):
                return f(v)
            case Maybe.Nothing:
                return False

    is_some_and = is_just_and

    def is_nothing(self) -> bool:
        """Returns true if the `Maybe` is a `Nothing` value."""
        return self is Maybe[T].Nothing

    is_none = is_nothing

    def ok_or[E](self, err: E) -> Result[T, E]:
        """
        Transforms the `Maybe[T]` into a `Result[T, E]`, mapping `Just(v)` to `Ok(v)`
        and `Nothing` to `Err(err)`.
        """
        from .result import Err, Ok

        match self:
            case Maybe.Just(v):
                return Ok(v)
            case Maybe.Nothing:
                return Err(err)

    def ok_or_else[E](self, err: Callable[[], E]) -> Result[T, E]:
        """
        Transforms the `Maybe[T]` into a `Result[T, E]`, mapping `Just(v)` to `Ok(v)`
        and `Nothing` to `Err(err())`.
        """
        from .result import Err, Ok

        match self:
            case Maybe.Just(v):
                return Ok(v)
            case Maybe.Nothing:
                return Err(err())

    def inspect(self, f: Callable[[T], Any]) -> Maybe[T]:
        """
        Calls a function with the contained value if `Just`.
        Returns the original `Maybe`.
        """
        match self:
            case Maybe.Just(v):
                f(v)
            case Maybe.Nothing:
                pass
        return self

    def expect(self, message: str) -> T:
        """
        Returns the contained `Just` value if the `Maybe` is a `Just` or raises a `MaybeUnwrapError`.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                raise MaybeUnwrapError(message)

    def unwrap(self) -> T:
        """
        Returns the contained value if the `Maybe` is a `Just` or raises a `MaybeUnwrapError`.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                raise MaybeUnwrapError('Called `Maybe.unwrap()` on a `Nothing` value')

    def unwrap_or(self, default: T) -> T:
        """
        Returns the contained `Just` value or a provided default.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                return default

    def unwrap_or_none(self) -> T | None:
        """
        Returns the contained `Just` value or `None`.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                return None

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        """
        Returns the contained `Just` value or computes it from a closure.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                return f()

    def unwrap_or_raise(self, exc: Callable[[], ExcType]) -> T:
        """
        Returns the contained `Just` value or raise the provided exception.
        """
        match self:
            case Maybe.Just(v):
                return v
            case Maybe.Nothing:
                raise exc()

    def map[U](self, f: Callable[[T], U]) -> Maybe[U]:
        """
        Maps a `Maybe[T]` to `Maybe[U]` by applying a function to a contained value (if `Just`)
        or returns `Nothing` (if `Nothing`).
        """
        match self:
            case Maybe.Just(v):
                return Maybe[U].Just(f(v))
            case Maybe.Nothing:
                return self

    async def map_async[U](self, f: Callable[[T], Awaitable[U]]) -> Maybe[U]:
        """
        Maps a `Maybe[T]` to `Maybe[U]` by applying a function to a contained value (if `Just`)
        or returns `Nothing` (if `Nothing`).
        """
        match self:
            case Maybe.Just(v):
                return Maybe[U].Just(await f(v))
            case Maybe.Nothing:
                return self

    def map_or[U](self, default: U, f: Callable[[T], U]) -> U:
        """
        Returns the provided default result (if `Nothing`), or applies a function
        to the contained value (if `Just`).
        """
        match self:
            case Maybe.Just(v):
                return f(v)
            case Maybe.Nothing:
                return default

    def map_or_else[U](self, default_f: Callable[[], U], f: Callable[[T], U]) -> U:
        """
        Computes a default function result (if `Nothing`), or applies a different function
        to the contained value (if `Just`).
        """
        match self:
            case Maybe.Just(v):
                return f(v)
            case Maybe.Nothing:
                return default_f()

    def and_[U](self, other: Maybe[U]) -> Maybe[U]:
        """
        Returns `Nothing` if the Maybe is `Nothing`, otherwise returns `other`.
        """
        match self:
            case Maybe.Just(_):
                return other
            case Maybe.Nothing:
                return self

    def and_then[U](self, f: Callable[[T], Maybe[U]]) -> Maybe[U]:
        """
        Returns `Nothing` if the Maybe is `Nothing`, otherwise calls `f` with the wrapped value and returns the result.
        Some languages call this operation flatmap.
        """
        match self:
            case Maybe.Just(v):
                return f(v)
            case Maybe.Nothing:
                return self

    async def and_then_async[U](self, f: Callable[[T], Awaitable[Maybe[U]]]) -> Maybe[U]:
        """
        Returns `Nothing` if the Maybe is `Nothing`, otherwise calls `f` with the wrapped value and returns the result.
        Some languages call this operation flatmap.
        """
        match self:
            case Maybe.Just(v):
                return await f(v)
            case Maybe.Nothing:
                return self

    def filter(self, predicate: Callable[[T], bool]) -> Maybe[T]:
        """
        Returns `Nothing` if the Maybe is `Nothing`, otherwise calls `predicate` with the wrapped value and returns:
        * `Just(t)` if predicate returns true (where t is the wrapped value), and
        * `Nothing` if predicate returns false.
        """
        match self:
            case Maybe.Just(v):
                if predicate(v):
                    return self
                return Maybe[T].Nothing
            case Maybe.Nothing:
                return self

    def or_(self, other: Maybe[T]) -> Maybe[T]:
        """
        Returns the Maybe if it contains a value, otherwise returns `other`.
        """
        match self:
            case Maybe.Just(_):
                return self
            case Maybe.Nothing:
                return other

    def or_else(self, f: Callable[[], Maybe[T]]) -> Maybe[T]:
        """
        Returns the Maybe if it contains a value, otherwise calls `f` and returns the result.
        """
        match self:
            case Maybe.Just(_):
                return self
            case Maybe.Nothing:
                return f()

    def xor(self, other: Maybe[T]) -> Maybe[T]:
        """
        Returns `Just` if exactly one of `self`, `other` is `Just`, otherwise returns `Nothing`.
        """
        match (self, other):
            case (Maybe.Just(_), Maybe.Nothing):
                return self
            case (Maybe.Nothing, Maybe.Just(_)):
                return other
            case _:
                return Maybe[T].Nothing

    def flatten(self) -> Maybe[T]:
        """
        Converts from `Maybe[Maybe[T]]` to `Maybe[T]`.
        """
        match self:
            case Maybe.Just(v):
                match v:
                    case Maybe.Just(_):
                        return v
                    case Maybe.Nothing:
                        return Maybe[T].Nothing
                    case _:
                        return self
            case Maybe.Nothing:
                return self

    def to_result(self) -> Result[Maybe[T], Any]:
        """
        Transposes a `Maybe` of a `Result` into an `Result` of a `Maybe`.

        `Nothing` will be mapped to `Ok(Nothing)`.
        `Just(Ok(_))` and `Just(Err(_))` will be mapped to `Ok(Just(_))` and `Err(_)`.
        """
        from .result import Err, Ok, Result

        match self:
            case Maybe.Just(v):
                match v:
                    case Result.Ok(x):
                        return Ok(Just(x))
                    case Result.Err(e):
                        return Err(e)
                    case _:
                        return Ok(self)

            case Maybe.Nothing:
                return Result[Maybe[T], Any].Ok(Nothing)


Just = Maybe.Just
Nothing = Maybe.Nothing


class MaybeUnwrapError(Exception):
    """
    Exception raised from ``.unwrap_<...>`` and ``.expect_<...>`` calls.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


def is_just[T](maybe: Maybe[T]) -> TypeGuard[Maybe[T].Just]:  # type: ignore
    """A typeguard to check if a Maybe is a Just

    Usage:
    >>> r: Maybe[int] = get_a_maybe()
    >>> if is_just(r):
    >>>     r   # r is of type Just[int]
    >>> elif is_nothing(r):
    >>>     r   # r is of type Nothing
    """
    return maybe.is_just()


def is_nothing[T](maybe: Maybe[T]) -> TypeGuard[Maybe[T].Nothing]:  # type: ignore
    """A typeguard to check if a Maybe is a Nothing

    Usage:
    >>> r: Maybe[int] = get_a_maybe()
    >>> if is_just(r):
    >>>     r   # r is of type Just[int]
    >>> elif is_nothing(r):
    >>>     r   # r is of type Nothing
    """
    return maybe.is_nothing()



from greyhorse.app.builders.loader import ModuleLoader
from greyhorse.app.entities.components import Component, ModuleComponent
from greyhorse.app.entities.module import Module
from greyhorse.app.schemas.components import ComponentConf, ModuleComponentConf, ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.result import Err, Ok, Result


class ModuleBuildError(Error):
    namespace = 'greyhorse.app.builders.module'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    Factory = ErrorCase(
        msg='Module factory error: "{path}", details: "{details}"', path=str, details=str
    )
    LoadError = ErrorCase(
        msg='Load error in component: "{path}", details: "{details}"', path=str, details=str
    )
    UnloadError = ErrorCase(
        msg='Unload error in component: "{path}", details: "{details}"', path=str, details=str
    )
    ComponentError = ErrorCase(
        msg='Component error in module: "{path}" "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )


class ComponentBuildError(Error):
    namespace = 'greyhorse.app.builders.component'

    Disabled = ErrorCase(msg='Component is disabled: "{path}" "{name}"', path=str, name=str)
    Factory = ErrorCase(
        msg='Component factory error: "{path}" "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )
    Submodule = ErrorCase(
        msg='Component submodule error: "{path}" "{name}", details: "{details}"',
        path=str,
        name=str,
        details=str,
    )


class ModuleBuilder:
    def __init__(self, conf: ModuleConf, path: str) -> None:
        self._conf = conf
        self._path = path

    def create_pass(self) -> Result[Module, ModuleBuildError]:
        if not self._conf.enabled:
            return ModuleBuildError.Disabled(path=self._path).to_result()

        logger.info('{path}: Module create'.format(path=self._path))

        if not (res := self._create_components()):
            return res

        components = res.unwrap()

        try:
            instance = Module(path=self._path, conf=self._conf, components=components)

        except Exception as e:
            error = ModuleBuildError.Factory(path=self._path, details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info('{path}: Module created successfully'.format(path=self._path))
        return Ok(instance)

    def destroy_pass(self) -> Result[None, ModuleBuildError]:
        if not self._conf.enabled:
            return ModuleBuildError.Disabled(path=self._path).to_result()

        logger.info('{path}: Module destroy'.format(path=self._path))

        if not (res := self._destroy_components()):
            return res

        logger.info('{path}: Module destroyed successfully'.format(path=self._path))
        return Ok()

    def _create_component(
        self, name: str, conf: ComponentConf
    ) -> Result[Component, ComponentBuildError]:
        if not conf.enabled:
            return ComponentBuildError.Disabled(path=self._path, name=name).to_result()

        logger.info('{path}: Component "{name}" creation'.format(path=self._path, name=name))

        try:
            instance = Component(name=name, path=self._path, conf=conf)

        except Exception as e:
            error = ComponentBuildError.Factory(path=self._path, name=name, details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component "{name}" created successfully'.format(path=self._path, name=name)
        )

        return Ok(instance)

    def _create_module_component(
        self, name: str, conf: ModuleComponentConf
    ) -> Result[ModuleComponent, ComponentBuildError]:
        if not conf.enabled:
            return ComponentBuildError.Disabled(path=self._path, name=name).to_result()

        logger.info(
            '{path}: Module component "{name}" create'.format(path=self._path, name=name)
        )

        if not (
            res := load_module(f'{self._path}.{name}', conf).map_err(
                lambda e: ComponentBuildError.Submodule(
                    path=self._path, name=name, details=e.message
                )
            )
        ):
            return res

        module = res.unwrap()

        try:
            instance = ModuleComponent(name=name, path=self._path, conf=conf, module=module)

        except Exception as e:
            error = ComponentBuildError.Factory(path=self._path, name=name, details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Module component "{name}" created successfully'.format(
                path=self._path, name=name
            )
        )

        return Ok(instance)

    def _create_components(self) -> Result[list[Component], ModuleBuildError]:
        result = []

        for name, conf in self._conf.components.items():
            match conf:
                case ModuleComponentConf() as conf:
                    match self._create_module_component(name, conf):
                        case Ok(component):
                            result.append(component)

                        case Err(e):
                            match e:
                                case ComponentBuildError.Disabled(_):
                                    logger.info(e.message)
                                case _:
                                    error = ModuleBuildError.ComponentError(
                                        path=self._path, name=name, details=e.message
                                    )
                                    logger.error(error.message)
                                    return error.to_result()

                case ComponentConf() as conf:
                    match self._create_component(name, conf):
                        case Ok(component):
                            result.append(component)

                        case Err(e):
                            match e:
                                case ComponentBuildError.Disabled(_):
                                    logger.info(e.message)
                                case _:
                                    error = ModuleBuildError.ComponentError(
                                        path=self._path, name=name, details=e.message
                                    )
                                    logger.error(error.message)
                                    return error.to_result()

        return Ok(result)

    def _destroy_components(self) -> Result[None, ModuleBuildError]:
        for name, conf in reversed(self._conf.components.items()):
            if not conf.enabled:
                return ComponentBuildError.Disabled(path=self._path, name=name).to_result()

            match conf:
                case ComponentConf():
                    pass

                case ModuleComponentConf() as conf:
                    logger.info(
                        '{path}: Module component "{name}" destroy'.format(
                            path=self._path, name=name
                        )
                    )

                    if not (
                        res := unload_module(f'{self._path}.{name}', conf).map_err(
                            lambda e: ModuleBuildError.UnloadError(
                                path=self._path, details=e.message
                            )
                        )
                    ):
                        return res

                    logger.info(
                        '{path}: Module component "{name}" destroyed successfully'.format(
                            path=self._path, name=name
                        )
                    )

        return Ok()


def load_module(path: str, conf: ModuleComponentConf) -> Result[Module, ModuleBuildError]:
    loader = ModuleLoader()

    if not (
        res := loader.load_pass(conf).map_err(
            lambda e: ModuleBuildError.LoadError(path=conf.path, details=e.message)
        )
    ):
        return res

    module_conf = res.unwrap()
    builder = ModuleBuilder(module_conf, path)

    if not (res := builder.create_pass()):
        return res

    module = res.unwrap()
    return Ok(module)


def unload_module(path: str, conf: ModuleComponentConf) -> Result[None, ModuleBuildError]:
    builder = ModuleBuilder(conf._conf, path)

    if not (res := builder.destroy_pass()):
        return res

    loader = ModuleLoader()

    return loader.unload_pass(conf).map_err(
        lambda e: ModuleBuildError.UnloadError(path=conf.path, details=e.message)
    )



from typing import Any, get_type_hints, override

from greyhorse.result import Ok, Result

from ..abc.collectors import Collector, MutCollector
from ..abc.controllers import Controller, ControllerError, OperatorMember


def operator(resource_type: type):
    def decorator(func: classmethod):
        class_name, method_name = func.__qualname__.split('.')
        hints = get_type_hints(func, include_extras=True)
        ret_type = hints.pop('return', None)
        func.__operator__ = OperatorMember(
            class_name, method_name, resource_type, method=func, params=hints, ret_type=ret_type
        )
        return func

    return decorator


class SyncController(Controller):
    @override
    def setup(self, collector: Collector[type, Any]) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    def teardown(self, collector: MutCollector[type, Any]) -> Result[bool, ControllerError]:
        return Ok(True)


class AsyncController(Controller):
    @override
    async def setup(self, collector: Collector[type, Any]) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    async def teardown(
        self, collector: MutCollector[type, Any]
    ) -> Result[bool, ControllerError]:
        return Ok(True)



from abc import ABC, abstractmethod

from greyhorse.utils.types import TypeWrapper


class Collector[K, T](TypeWrapper[K, T], ABC):
    @abstractmethod
    def add(self, key: K, instance: T) -> bool: ...


class MutCollector[K, T](Collector[K, T], ABC):
    @abstractmethod
    def remove(self, key: K, instance: T | None = None) -> bool: ...



from datetime import datetime, timedelta
from unittest import mock
from unittest.mock import call

import pytest
from greyhorse.data.cache.base import CacheData, ModelCacheOperator
from greyhorse.data.cache.method import MethodCache


@pytest.mark.asyncio
async def test_keys() -> None:
    operator_mock = mock.AsyncMock(spec=ModelCacheOperator)
    method_cache = MethodCache(operator_mock, timedelta(seconds=3))

    operator_mock.get_model_cache_key.return_value = 'greyhorse.models.test'
    operator_mock.get_cache_id.return_value = '123'
    operator_mock.get_cache_key.return_value = 'greyhorse.models.test:1'

    assert method_cache.get_model_cache_key() == 'greyhorse.models.test'
    assert method_cache.get_cache_id(123) == '123'
    assert method_cache.get_cache_key(1) == 'greyhorse.models.test:1'

    operator_mock.get_model_cache_key.assert_called_once()
    operator_mock.get_cache_id.assert_called_once_with(123)
    operator_mock.get_cache_key.assert_called_once_with(1)
    operator_mock.get_model_cache_key.reset_mock()
    operator_mock.get_cache_id.reset_mock()
    operator_mock.get_cache_key.reset_mock()


@pytest.mark.asyncio
async def test_put() -> None:
    operator_mock = mock.AsyncMock(spec=ModelCacheOperator)
    method_cache = MethodCache(operator_mock, timedelta(seconds=3))
    args = {'a': 1, 'b': 12.34, 'c': 'asdf', 'd': datetime.now()}

    operator_mock.cache_one.return_value = True
    assert await method_cache.put('method', None, None)
    operator_mock.cache_one.assert_called_once_with(
        CacheData(cache_id=MethodCache.cache_key_for('method', None), data=None),
        ttl=timedelta(seconds=3),
    )
    operator_mock.cache_one.reset_mock()

    operator_mock.cache_one.return_value = True
    assert await method_cache.put('method', args, 123)
    operator_mock.cache_one.assert_called_once_with(
        CacheData(cache_id=MethodCache.cache_key_for('method', args), data=123),
        ttl=timedelta(seconds=3),
    )
    operator_mock.cache_one.reset_mock()


@pytest.mark.asyncio
async def test_get() -> None:
    operator_mock = mock.AsyncMock(spec=ModelCacheOperator)
    method_cache = MethodCache(operator_mock)
    args = {'a': 1, 'b': 12.34, 'c': 'asdf', 'd': datetime.now()}

    operator_mock.get_cache_key.return_value = 'greyhorse.models.test:1'

    operator_mock.load_one.return_value = (False, None)
    exists, data = await method_cache.get('method', args)
    assert not exists
    assert not data
    operator_mock.get_cache_key.assert_called_once_with(
        MethodCache.cache_key_for('method', args)
    )
    operator_mock.load_one.assert_called_once_with('greyhorse.models.test:1')
    operator_mock.get_cache_key.reset_mock()
    operator_mock.load_one.reset_mock()

    operator_mock.load_one.return_value = (True, 123)
    exists, data = await method_cache.get('method', args)
    assert exists
    assert data == 123
    operator_mock.get_cache_key.assert_called_once_with(
        MethodCache.cache_key_for('method', args)
    )
    operator_mock.load_one.assert_called_once_with('greyhorse.models.test:1')
    operator_mock.get_cache_key.reset_mock()
    operator_mock.load_one.reset_mock()


@pytest.mark.asyncio
async def test_drop() -> None:
    operator_mock = mock.AsyncMock(spec=ModelCacheOperator)
    method_cache = MethodCache(operator_mock)
    args = {'a': 1, 'b': 12.34, 'c': 'asdf', 'd': datetime.now()}

    operator_mock.get_cache_key.return_value = 'greyhorse.models.test:1'

    operator_mock.drop_one.return_value = True
    assert await method_cache.drop('method', args)
    operator_mock.get_cache_key.assert_called_once_with(
        MethodCache.cache_key_for('method', args)
    )
    operator_mock.drop_one.assert_called_once_with('greyhorse.models.test:1')
    operator_mock.get_cache_key.reset_mock()
    operator_mock.drop_one.reset_mock()


@pytest.mark.asyncio
async def test_drop_all() -> None:
    operator_mock = mock.AsyncMock(spec=ModelCacheOperator)
    method_cache = MethodCache(operator_mock)

    operator_mock.drop_all.return_value = 1
    assert await method_cache.drop_all() == 1
    operator_mock.drop_all.assert_called_once()
    operator_mock.drop_all.reset_mock()

    assert await method_cache.drop_all(['1', '2']) == 2
    operator_mock.drop_all.assert_has_calls([call('1'), call('2')], any_order=True)
    operator_mock.drop_all.reset_mock()



from importlib.resources import Package, as_file, files
from pathlib import Path

import tomlkit

from greyhorse.utils.dicts import build_dotted_keys_from_dict


class StaticTranslator:
    def __init__(self) -> None:
        self._data = {}
        self._defaults = {}

    def load_string(self, content: str, namespace: str) -> None:
        parsed = tomlkit.parse(content)
        self._defaults[namespace] = parsed.pop('default', None)
        self._data |= build_dotted_keys_from_dict(parsed, root_key=namespace)

    def load_file(self, filename: str, namespace: str | None = None) -> None:
        filename = Path(filename)
        namespace = namespace if namespace else filename.stem
        self.load_string(filename.read_text('utf-8'), namespace)

    def load_package(
        self, package: Package, filename: str, namespace: str | None = None
    ) -> None:
        filename = Path(filename)
        namespace = namespace if namespace else filename.stem
        source = files(package).joinpath(filename)
        with as_file(source) as file:
            self.load_string(file.read_text(), namespace)

    def unload(self, namespace: str) -> None:
        self._defaults.pop(namespace, None)
        keys_to_remove = [k for k, v in self._data.items() if k.startswith(f'{namespace}.')]
        for k in keys_to_remove:
            self._data.pop(k)

    def size(self) -> int:
        return len(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def set_default_lang(self, namespace: str, lang: str) -> bool:
        if namespace not in self._defaults:
            return False
        self._defaults[namespace] = lang
        return True

    def namespaces(self) -> frozenset[str]:
        return frozenset(self._defaults.keys())

    def __call__(
        self, key: str, /, lang: str | None = None, default: str = '', **kwargs
    ) -> str:
        if len(self._defaults) == 1:
            ns = next(iter(self._defaults.keys()))
        else:
            ns = key.find('.')
            ns = key[0:ns] if ns != -1 else None

        tries = []

        if lang:
            tries += [f'{key}.{lang}']
            if ns and ns in self._defaults:
                tries += [f'{ns}.{key}.{lang}']
                default_lang = self._defaults.get(ns)
                if default_lang != lang:
                    tries += [f'{key}.{default_lang}', f'{ns}.{key}.{default_lang}']

        tries += [key]
        if len(self._defaults) == 1:
            tries += [f'{ns}.{key}']
        if not lang and ns and ns in self._defaults:
            default_lang = self._defaults.get(ns)
            tries += [f'{key}.{default_lang}', f'{ns}.{key}.{default_lang}']

        for t in tries:
            if res := self._data.get(t):
                return res.format(**kwargs) if kwargs else res

        return default.format(**kwargs) if kwargs else default



from collections.abc import Callable, Iterable, Mapping
from typing import Any


def dict_values_to_str(data: dict):
    result = dict()

    for k, v in data.items():
        if isinstance(v, str):
            pass
        elif isinstance(v, dict):
            v = dict_values_to_str(v)
        elif isinstance(v, list):
            v = [str(i) for i in v]
        else:
            v = str(v)

        result[str(k)] = v

    return result


def build_dict_from_dotted_keys(
    iterable: Iterable, key_getter: Callable[[Any], Any], value_getter: Callable[[Any], Any]
) -> Mapping[str, Any]:
    result = dict()

    for obj in iterable:
        keys = key_getter(obj).split('.')
        cur_dict = result

        for key in keys[:-1]:
            if key not in cur_dict:
                cur_dict[key] = {}

            if not isinstance(cur_dict[key], dict):
                cur_dict[key] = {'.': cur_dict[key]}
            cur_dict = cur_dict[key]

        cur_dict[keys[-1]] = value_getter(obj)

    return result


def build_dotted_keys_from_dict(
    dict_: Mapping[str, Any], root_key: str | None = None
) -> Mapping[str, Any]:
    def traverse(key_stack: list[str], values: Mapping[str, Any]) -> Mapping[str, Any]:
        result = dict()

        for k, v in values.items():
            if isinstance(v, dict):
                result.update(traverse([*key_stack, k], v))
            elif isinstance(v, list):
                for i in v:
                    i = traverse([*key_stack, k], i)
                    if isinstance(i, dict):
                        result |= i
            else:
                result['.'.join([*key_stack, k])] = v

        return result

    return traverse([root_key] if root_key else [], dict_)


def obj_dict_to_str_dict(data: dict, value_getter: Callable[[Any], Any]) -> Mapping[str, Any]:
    def traverse(values: Mapping[str, Any]) -> Mapping[str, Any]:
        result = dict()

        for k, v in values.items():
            if isinstance(v, dict):
                result[k] = traverse(v)
            elif v is None:
                pass
            else:
                result[k] = value_getter(v)

        return result

    return traverse(data)



import contextlib
from dataclasses import dataclass
from datetime import datetime
from unittest.mock import MagicMock, Mock

import pytest
from greyhorse.app.contexts import AsyncContext, ContextBuilder, SyncContext


def test_sync_context_scalar() -> None:
    ctx_builder = ContextBuilder[SyncContext, int](lambda: 123)

    mock_context = MagicMock(spec=contextlib.AbstractContextManager)

    @contextlib.contextmanager
    def mock_fn():
        with mock_context:
            yield

    mock_finalizer = Mock()

    ctx_builder.add_context(mock_fn)
    ctx_builder.add_finalizer(mock_finalizer)

    ctx = ctx_builder.build()

    with ctx as number:
        assert number == 123

    mock_context.__enter__.assert_called_once()
    mock_context.__exit__.assert_called_once()
    mock_finalizer.assert_called_once()


@dataclass
class ContextData:
    id: int
    name: str
    timestamp: datetime


def test_sync_context_complex() -> None:
    ctx_builder = ContextBuilder[SyncContext, ContextData](ContextData)

    mock_context = MagicMock(spec=contextlib.AbstractContextManager)

    @contextlib.contextmanager
    def mock_fn():
        with mock_context:
            yield

    mock_finalizer = Mock()

    ctx_builder.add_context(mock_fn)
    ctx_builder.add_finalizer(mock_finalizer)
    ctx_builder.add_param('id', 123)
    ctx_builder.add_param('name', lambda: 'name')
    ctx_builder.add_param('timestamp', datetime.now)

    ctx = ctx_builder.build()

    with ctx as data:
        assert data.id == 123
        assert data.name == 'name'
        assert data.timestamp.replace(second=0, microsecond=0) == datetime.now().replace(
            second=0, microsecond=0
        )

    mock_context.__enter__.assert_called_once()
    mock_context.__exit__.assert_called_once()
    mock_finalizer.assert_called_once()


@pytest.mark.asyncio
async def test_async_context_scalar() -> None:
    ctx_builder = ContextBuilder[AsyncContext, int](lambda: 123)

    mock_context = MagicMock(spec=contextlib.AbstractAsyncContextManager)

    @contextlib.asynccontextmanager
    async def mock_fn():
        async with mock_context:
            yield

    mock_finalizer = Mock()

    ctx_builder.add_context(mock_fn)
    ctx_builder.add_finalizer(mock_finalizer)

    ctx = ctx_builder.build()

    async with ctx as number:
        assert number == 123

    mock_context.__aenter__.assert_called_once()
    mock_context.__aexit__.assert_called_once()
    mock_finalizer.assert_called_once()


@pytest.mark.asyncio
async def test_async_context_complex() -> None:
    ctx_builder = ContextBuilder[AsyncContext, ContextData](ContextData)

    mock_context = MagicMock(spec=contextlib.AbstractAsyncContextManager)

    @contextlib.asynccontextmanager
    async def mock_fn():
        async with mock_context:
            yield

    async def _get_name() -> str:
        return 'name'

    mock_finalizer = Mock()

    ctx_builder.add_context(mock_fn)
    ctx_builder.add_finalizer(mock_finalizer)
    ctx_builder.add_param('id', 123)
    ctx_builder.add_param('name', _get_name)
    ctx_builder.add_param('timestamp', datetime.now)

    ctx = ctx_builder.build()

    async with ctx as data:
        assert data.id == 123
        assert data.name == 'name'
        assert data.timestamp.replace(second=0, microsecond=0) == datetime.now().replace(
            second=0, microsecond=0
        )

    mock_context.__aenter__.assert_called_once()
    mock_context.__aexit__.assert_called_once()
    mock_finalizer.assert_called_once()



from collections.abc import Mapping
from typing import Any, Protocol


class Serializer(Protocol):
    def serialize(self, data: Any | None = None) -> bytes: ...


class Deserializer(Protocol):
    def deserialize(self, data: bytes) -> Any | None: ...


class ModelSerializer(Protocol):
    serializer: Serializer

    def serialize(self, only_fields: list[str] | None = None) -> bytes:
        return self.serializer.serialize(self.get_serializable_values(only_fields))

    def get_serializable_values(
        self, only_fields: list[str] | None = None
    ) -> Mapping[str, Any] | None: ...



from collections.abc import Callable
from copy import deepcopy
from typing import Any, override

from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import (
    BorrowError,
    BorrowMutError,
    ForwardError,
    ForwardProvider,
    MutProvider,
    SharedProvider,
)
from greyhorse.app.contexts import (
    AsyncContext,
    AsyncMutContext,
    ContextBuilder,
    SyncContext,
    SyncMutContext,
)
from greyhorse.maybe import Maybe, Nothing
from greyhorse.result import Ok, Result


class _BasicRefBox:
    __slots__ = ('_shared_counter', '_acq_counter')

    allow_borrow_when_acquired = False
    allow_acq_when_borrowed = False
    allow_multiple_acquisition = False

    def __init__(self) -> None:
        self._shared_counter = 0
        self._acq_counter = 0

    def _borrow[T](self, value: T) -> Result[T, BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=type(value).__name__).to_result()
        self._shared_counter += 1
        return Ok(value)

    def _borrow_mut[T](self, value: T) -> Result[T, BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(name=type(value).__name__).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(name=type(value).__name__).to_result()
        self._acq_counter += 1
        return Ok(value)

    @staticmethod
    def _ensure_maybe(value):
        if not isinstance(value, Maybe):
            value = Maybe(value)
        return value


class SharedRefBox[T](_BasicRefBox, SharedProvider[T]):
    __slots__ = ('_factory', '_copy_maker')

    def __init__(
        self, factory: Callable[[], Maybe[T] | T], copy_maker: Callable[[T], T] = deepcopy
    ) -> None:
        super().__init__()
        self._factory = factory
        self._copy_maker = copy_maker

    @override
    def borrow(self) -> Result[T, BorrowError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type.__name__).to_result(),
                self._borrow,
            )
            .map(self._copy_maker)
        )

    @override
    def reclaim(self, instance: T) -> None:
        self._shared_counter -= 1
        del instance


class MutRefBox[T](_BasicRefBox, MutProvider[T]):
    __slots__ = ('_factory', '_copy_maker')

    def __init__(
        self, factory: Callable[[], Maybe[T] | T], copy_maker: Callable[[T], T] = deepcopy
    ) -> None:
        super().__init__()
        self._factory = factory
        self._copy_maker = copy_maker

    @override
    def acquire(self) -> Result[T, BorrowMutError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type.__name__).to_result(),
                self._borrow_mut,
            )
            .map(self._copy_maker)
        )

    @override
    def release(self, instance: T) -> None:
        self._acq_counter -= 1
        del instance


class OwnerRefBox[TS, TM](_BasicRefBox, SharedProvider[TS], MutProvider[TM]):
    __slots__ = ('_factory', '_mut_factory', '_copy_maker', '_mut_copy_maker')

    def __init__(
        self,
        factory: Callable[[], Maybe[TS] | TS],
        mut_factory: Callable[[], Maybe[TM] | TM],
        copy_maker: Callable[[TS], TS] = deepcopy,
        mut_copy_maker: Callable[[TM], TM] = deepcopy,
    ) -> None:
        super().__init__()
        self._factory = factory
        self._mut_factory = mut_factory
        self._copy_maker = copy_maker
        self._mut_copy_maker = mut_copy_maker

    @override
    def borrow(self) -> Result[TS, BorrowError]:
        return (
            self._ensure_maybe(self._factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type[0].__name__).to_result(),
                self._borrow,
            )
            .map(self._copy_maker)
        )

    @override
    def reclaim(self, instance: TS) -> None:
        self._shared_counter -= 1
        del instance

    @override
    def acquire(self) -> Result[TM, BorrowMutError]:
        return (
            self._ensure_maybe(self._mut_factory())
            .map_or_else(
                lambda: BorrowError.Empty(name=self.wrapped_type[1].__name__).to_result(),
                self._borrow_mut,
            )
            .map(self._mut_copy_maker)
        )

    @override
    def release(self, instance: TM) -> None:
        self._acq_counter -= 1
        del instance


class SharedCtxRefBox[T](_BasicRefBox, SharedProvider[T]):
    __slots__ = ('_kind', '_params', '_factory')

    def __init__(
        self, kind: type[SyncContext[T] | AsyncContext[T]], factory: Callable[[], T], **params
    ) -> None:
        super().__init__()
        self._kind = kind
        self._params = params
        self._factory = factory

    @override
    def borrow(self) -> Result[SyncContext[T] | AsyncContext[T], BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=self.wrapped_type.__name__).to_result()
        self._shared_counter += 1

        ctx_builder = ContextBuilder[self._kind, self.wrapped_type](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def reclaim(self, instance: SyncContext[T] | AsyncContext[T]) -> None:
        self._shared_counter -= 1
        del instance


class MutCtxRefBox[T](_BasicRefBox, MutProvider[T]):
    __slots__ = ('_mut_kind', '_params', '_factory')

    def __init__(
        self,
        mut_kind: type[SyncMutContext[T] | AsyncMutContext[T]],
        factory: Callable[[], T],
        **params,
    ) -> None:
        super().__init__()
        self._mut_kind = mut_kind
        self._params = params
        self._factory = factory

    @override
    def acquire(self) -> Result[SyncMutContext[T] | AsyncMutContext[T], BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(name=self.wrapped_type.__name__).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(
                name=self.wrapped_type.__name__
            ).to_result()
        self._acq_counter += 1

        ctx_builder = ContextBuilder[self._mut_kind, self.wrapped_type](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def release(self, instance: SyncMutContext[T] | AsyncMutContext[T]) -> None:
        self._acq_counter -= 1
        del instance


class OwnerCtxRefBox[TS, TM](_BasicRefBox, SharedProvider[TS], MutProvider[TM]):
    __slots__ = ('_kind', '_mut_kind', '_params', '_mut_params', '_factory', '_mut_factory')

    def __init__(
        self,
        kind: type[SyncContext[TS] | AsyncContext[TS]],
        mut_kind: type[SyncMutContext[TM] | AsyncMutContext[TM]],
        factory: Callable[[], TS],
        mut_factory: Callable[[], TM],
        params: dict[str, Any] | None = None,
        mut_params: dict[str, Any] | None = None,
    ) -> None:
        super().__init__()
        self._kind = kind
        self._mut_kind = mut_kind
        self._params = params or {}
        self._mut_params = mut_params or {}
        self._factory = factory
        self._mut_factory = mut_factory

    @override
    def borrow(self) -> Result[SyncContext[TS] | AsyncContext[TS], BorrowError]:
        if not self.allow_borrow_when_acquired and self._acq_counter > 0:
            return BorrowError.BorrowedAsMutable(name=self.wrapped_type[0].__name__).to_result()
        self._shared_counter += 1

        ctx_builder = ContextBuilder[self._kind, self.wrapped_type[0]](
            self._factory, **self._params
        )
        return Ok(ctx_builder.build())

    @override
    def reclaim(self, instance: SyncContext[TS] | AsyncContext[TS]) -> None:
        self._shared_counter -= 1
        del instance

    @override
    def acquire(self) -> Result[SyncMutContext[TM] | AsyncMutContext[TM], BorrowMutError]:
        if not self.allow_multiple_acquisition and self._acq_counter > 0:
            return BorrowMutError.AlreadyBorrowed(
                name=self.wrapped_type[1].__name__
            ).to_result()
        if not self.allow_acq_when_borrowed and self._shared_counter > 0:
            return BorrowMutError.BorrowedAsImmutable(
                name=self.wrapped_type[1].__name__
            ).to_result()
        self._acq_counter += 1

        ctx_builder = ContextBuilder[self._mut_kind, self.wrapped_type[1]](
            self._mut_factory, **self._mut_params
        )
        return Ok(ctx_builder.build())

    @override
    def release(self, instance: SyncMutContext[TM] | AsyncMutContext[TM]) -> None:
        self._acq_counter -= 1
        del instance


class ForwardBox[T](Operator[T], ForwardProvider[T]):
    __slots__ = ('_value',)

    def __init__(self, value: T | None = None) -> None:
        self._value = Maybe(value)

    @override
    def accept(self, value: T) -> bool:
        if self._value.is_just():
            return False
        self._value = Maybe(value)
        return True

    @override
    def revoke(self) -> Maybe[T]:
        value, self._value = self._value, Nothing
        return value

    @override
    def take(self) -> Result[T, ForwardError]:
        value, self._value = self._value, Nothing
        return value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        del instance

    @override
    def __bool__(self) -> bool:
        return self._value.is_just()


class PermanentForwardBox[T](ForwardBox[T]):
    @override
    def take(self) -> Result[T, ForwardError]:
        return self._value.map(Ok).unwrap_or(
            ForwardError.Empty(name=self.wrapped_type.__name__).to_result()
        )

    @override
    def drop(self, instance: T) -> None:
        pass
